%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (CHEM-E7225) ADVANCED PROCESS CONTROL D
%   #02 - Task 4 (Nonlinear optimisation)
%
%   Substitute the files ThreeTank.m and rk4fnc.m with your solutions
%   from HW1, then complete this script to perform parameter fitting 
%   for the benchmark Three-Tank system described in Task 4.
%
%   Find and replace each '???' with the correct code. 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all; clear; clc

addpath('~/path/to/casadi')   % <~ Change to the appropriate path in your computer
import casadi.*

opti = casadi.Opti();       % Creates an Opti Stack structure

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define the initial value problem (IVP)

% Experiment data (including (dT,K), assumed to be given)
load('ThreeTank_Data.mat', 'u_data', 'y_data', 'dT', 'K')

% Model parameters
S_T = ???;  S_V = ???;  g = ???;

a_V = opti.variable();  % The parameters (a_V,a_0) are not given
a_0 = opti.variable();  %  values, but defined as decision variables

px = [S_T; S_V; a_V; a_0; g];

% Initial state
x0 = ???;

% Integrator function (F)
F_rk4 = @(x,u) rk4fnc(@ThreeTank, dT, x, u, px);

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Define the optimisation problem

% Defines auxiliary variables for storing each (x_k,y_k)
% (Obs.: In Opti-Stack, we use opti.variable() instead of MX.sym() 
%        but they are equivalent, i.e., they both create MX objects)
x = opti.variable(3, K); 
y = opti.variable(2, K);

x(:,1) = x0;              % Initial state

% Loops through (u_data,y_data) to build the cost function J (sumJ)
sumJ = 0;
for k = 1:K-1
    x(:,k+1) = ???;       % <~ Fill with the expressions in Eq. (3c)-(3d)
    y(:,k)   = ???;       %    using  x(:,k)  and  u_data(:,k)

    sumJ = sumJ + ???;    % <~ Fill with the term inside the sum in Eq. (3a)
                          %    using  y(:,k)  and  y_data(:,k)
end
y(:,K) = ???;             % <~ Fill with the expression for y(K) = g(x(K))
sumJ = sumJ + ???;        % <~ Fill with the cost for the K-th output

% Creates the optimisation problem
opti.minimize( sumJ );
opti.subject_to( ??? );   % <~ Fill with the equivalent constraints
opti.subject_to( ??? );   %    for (a_v,a_0) given in Eq. (3b)

opti.solver('ipopt');
opti.solve();

display( "The solution is (a_V, a_0) = ("+opti.value(a_V)+", "+opti.value(a_0)+"). "  )


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plotting all the simulation results
%
%  NOTE: You don't need to complete this part, but feel free to tweak it
%        to customize the plotting more according to your style. 

x_data = full( opti.value(x) );
t_data = (0:K-1)*dT;

ms = 3.0; lw = 1.0; fs = 14;
figure(1), clf; 
for i = 1:4
  for j = 1:2
    subplot(2,3,j+(j==2)), 
      stairs(t_data, u_data(j,:), 'k-')
      ylabel("$u^{{data}}_"+j+"(t)$ [ml/s]",'Interpreter','Latex'); ylim([0,140])
      set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
      box off
  end

  subplot(2,3,4), plot(t_data,y_data(1,:)', 'r.', 'markersize', ms);
  subplot(2,3,6), plot(t_data,y_data(2,:)', 'r.', 'markersize', ms);

  for j = 1:3
    subplot(2,3,3+j), hold on 
      stairs(t_data,x_data(j,:)', 'k-'); 
      ylabel("$x_"+j+"(t)$ [cm]",'Interpreter','Latex'); ylim([0,60])
      xlabel('Time  - $t$ [s]','Interpreter','Latex')
      set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
      box off
  end
end
subplot(2,3,6)
legend(["Measurements ($y^{{data}}$)", "Fitted model"], ...
        'Location','northwest','Interpreter','Latex', ...
        'FontSize',10, 'Box', 'off')
